/* vim: set expandtab tabstop=4 shiftwidth=4: */
#if HAVE_CONFIG_H
#  include "config.h"
#endif

#include "includes.h"

static int        clear_close __P((int *));
static void       daemonize __P((void));
static void       do_bat __P((int));
static char *     get_program_name __P((char **));
static int        initialize __P((int, char **));
static int        initialize_options __P((int, char **));
static void       log_message __P((int, const char *, ...));
static int        tcp_accept __P((int));
static int        tcp_listen __P((unsigned short));
static void       quit __P((int));
static void       usage __P((int));
static RETSIGTYPE wait_child_process __P((int));
static pid_t      xfork __P((void));

enum log_type {
    LOG_TYPE_FILE,
    LOG_TYPE_TTY,
    LOG_TYPE_SYSLOG
};

static struct {
    int            daemon;
    int            inetd;
    enum log_type  log_type;
    char *         log_file;
    unsigned short port;
    int            thread;
} ccho_options = {
    0,            // daemon
    0,            // inetd
    LOG_TYPE_TTY, // log_type
    NULL,         // log_file
    7,            // port
    0             // thread
};

static int ccho_accepted_fd = 0;
static int ccho_listen_fd = 0;

static char *program_name = PACKAGE;

#define tcp_set_addr(addr, host, port) { \
  memset((voidptr)&(addr), 0, sizeof(addr)); \
  (addr).sin_family = AF_INET; \
  (addr).sin_port = (port); \
  (addr).sin_addr.s_addr = (host); \
}

int
main(argc, argv)
    int argc;
    char **argv;
{
    if (initialize(argc, argv) < 0) {
        quit(EXIT_FAILURE);
    }

    if (ccho_options.inetd) {
        bat(stdin, stdout);
        quit(EXIT_SUCCESS);
    }

    if (ccho_options.daemon) {
        daemonize();
    }

    ccho_listen_fd = tcp_listen(ccho_options.port);
    if (ccho_listen_fd < 0) {
        quit(EXIT_FAILURE);
    }

    while (1) {
        pid_t pid;

        ccho_accepted_fd = tcp_accept(ccho_listen_fd);
        if (ccho_accepted_fd < 0) {
            continue;
        }

        pid = fork();
        if (pid == 0) {
            clear_close(&ccho_listen_fd);
            do_bat(ccho_accepted_fd);
            quit(EXIT_SUCCESS);
        }

        clear_close(&ccho_accepted_fd);
    }
}

static int
clear_close(fd)
    int *fd;
{
    int tmp_fd;

    if (*fd <= 0) {
        return 0;
    }

    tmp_fd = *fd;
    *fd = 0;
    return close(tmp_fd);
}

static void
daemonize()
{
    int i;
    pid_t pid;

    pid = xfork();
    if (pid != 0) {
        quit(EXIT_SUCCESS);
    }

    setsid();

    pid = xfork();
    if (pid != 0) {
        quit(EXIT_SUCCESS);
    }

    chdir("/");
    umask(0);
    for (i = 0; i <= 2; i++) {
        close(i);
    }
}

static void
do_bat(fd)
    int fd;
{
    char buffer[BUFSIZ];
    int n;
    FILE *instream, *outstream;

    instream = fdopen(fd, "r");
    setlinebuf(instream);

    outstream = fdopen(fd, "w");
    setlinebuf(outstream);

    bat(instream, outstream);
}

static char *
get_program_name(argv)
    char **argv;
{
    char *tmp_program_name;

    tmp_program_name = basename(argv[0]);
    if (!tmp_program_name) {
        return PACKAGE;
    }

    tmp_program_name = strdup(tmp_program_name);
    if (!tmp_program_name) {
        return PACKAGE;
    }

    return tmp_program_name;
}

static int
initialize(argc, argv)
    int argc;
    char **argv;
{
    program_name = get_program_name(argv);
    if (strncmp(program_name, "in.", 3) == 0) {
        ccho_options.daemon = 0;
        ccho_options.inetd  = 1;
    }

    signal(SIGHUP, SIG_IGN);
    signal(SIGTERM, quit);
    signal(SIGCHLD, wait_child_process);

    if (initialize_options(argc, argv) < 0) {
        usage(EXIT_FAILURE);
    }

    return 0;
}

static int
initialize_options(argc, argv)
    int argc;
    char **argv;
{
    int c;
    static struct option longopts[] = {
        {"daemon",  no_argument,       NULL, 'd'},
        {"help",    no_argument,       NULL, 'h'},
        {"inetd",   no_argument,       NULL, 'i'},
        {"port",    required_argument, NULL, 'p'},
        {"syslog",  no_argument,       NULL, 's'},
        {"version", no_argument,       NULL, 'v'},
        {NULL,      0,                 NULL, 0}
    };

    while ((c = getopt_long(argc, argv, "dhip:sv", longopts, NULL)) != -1) {
        switch (c) {
        case 'd':
            ccho_options.daemon = 1;
            ccho_options.inetd  = 0;
            break;
        case 'h':
            usage(EXIT_SUCCESS);
            break;
        case 'i':
            ccho_options.daemon = 0;
            ccho_options.inetd  = 1;
            break;
        case 'p':
            {
                unsigned short port;
                port = (unsigned short)strtol(optarg, (char **)NULL, 0);
                if (port <= 0) {
                    return -1;
                }
                ccho_options.port = port;
            }
            break;
        case 's':
            ccho_options.log_type = LOG_TYPE_SYSLOG;
            break;
        case 'v':
            fprintf(stderr, " - %s Version "VERSION"\n", program_name);
            quit(EXIT_SUCCESS);
            break;
        default:
            usage(EXIT_FAILURE);
            break;
        }
    }

    return 0;
}

static void
#if STDC_HEADERS
log_message(int priority, const char *format, ...)
#else
log_message(priority, message, va_alist)
    int priority;
    const char *format;
    va_dcl;
#endif
{
    va_list ap;

    va_init(ap, format);

    switch (ccho_options.log_type) {
    case LOG_TYPE_FILE:
        break;
    case LOG_TYPE_TTY:
        vfprintf(stderr, format, ap);
        break;
    case LOG_TYPE_SYSLOG:
        openlog(NULL, LOG_PID, LOG_DAEMON);
        vsyslog(priority, format, ap);
        closelog();
        break;
    }

    va_end(ap);
}

static int
tcp_accept(listen_fd)
    int listen_fd;
{
    int accepted_fd;
    struct sockaddr_in addr;
    socklen_t l;

    l = sizeof(addr);
    return accept(listen_fd, (struct sockaddr *)&addr, &l);
}

static int
tcp_listen(port)
    unsigned short port;
{
    int fd = 0;
    struct sockaddr_in addr;

    fd = socket(AF_INET, SOCK_STREAM, 0);
    if (fd == -1) {
        goto failure_of_tcp_listen;
    }

    tcp_set_addr(addr, htonl(INADDR_ANY), htons(port));
    if (bind(fd, (struct sockaddr *)&addr, sizeof(addr)) == -1) {
        goto failure_of_tcp_listen;
    }

    if (listen(fd, 5) == -1) {
        goto failure_of_tcp_listen;
    }

    return fd;

failure_of_tcp_listen:
    if (fd) {
        close(fd);
    }
    return -1;
}

static void
quit(result)
    int result;
{
    clear_close(&ccho_accepted_fd);
    clear_close(&ccho_listen_fd);

    if (result != 0) {
        char error_message[BUFSIZ];
        strerror_r(errno, error_message, BUFSIZ);
        log_message(LOG_ERR, "%s\n", error_message);
    }

    exit(result);
}

static void
usage(result)
    int result;
{
    const char message[] = "usage: %s [options]...\n"
        "  -d,  --daemon     work as a daemon\n"
        "  -h,  --help       show this message\n"
        "  -i,  --inetd      work as an inetd daemon\n"
        "  -p,  --port=PORT  indicate port number\n"
        "  -s,  --syslog     use syslog as a logging system\n"
        "  -v,  --version    show version information\n"
        "";
    fprintf(stderr, message, program_name);
    quit(result);
}

static RETSIGTYPE
wait_child_process(dummy)
    int dummy;
{
    int status;

    while (wait(&status) > 0)
        ;

    signal(SIGCHLD, wait_child_process);
}

static pid_t
xfork()
{
    pid_t pid;

    pid = fork();
    if (pid < 0) {
        quit(EXIT_FAILURE);
    }

    return pid;
}
